"""Book's file."""
import os
import re
import urllib.parse
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, List, Optional, Tuple

from goodreads_export.clean_file_name import clean_file_name


@dataclass
class BookFile:  # pylint: disable=too-many-instance-attributes
    """Book's markdown file.

    On init extract fields from `content` - override other parameters.
    To re-parse the `content` call `parse()`.
    `render()` generate `content` from fields.
    """

    title: str
    folder: Path
    content: Optional[str] = None
    file_name: Optional[str] = None
    author: Optional[str] = None
    book_id: Optional[str] = None
    tags: Optional[List[str]] = None
    rating: Optional[int] = None
    isbn: Optional[int] = None
    isbn13: Optional[int] = None
    review: Optional[int] = None
    series: Optional[List[str]] = None

    _file_name: Optional[str] = field(repr=False, init=False)
    _goodreads_link_pattern: re.Pattern[str] = field(
        repr=False,
        init=False,
        default=re.compile(
            r"\[\[([^]]+)\]\](: \[([^]]*)\]\(https://www\.goodreads\.com/book/show/(\d+)\))"
        ),
    )
    _series_link_pattern: re.Pattern[str] = field(
        repr=False,
        init=False,
        default=re.compile(r"\[\[[^-]* - ([^-]*) - series\]\]"),
    )

    def __post_init__(self) -> None:
        """Extract fields from content."""
        self.parse()  # we do not run parse on content assign during __init__()
        if self.content is None:
            self.render()
        if self.tags is None:
            self.tags = []
        if self.series is None:
            self.series = []

    def parse(self) -> None:
        """Parse markdown file content."""
        if self.content is not None:
            if (series_match := self._goodreads_link_pattern.search(self.content)) is not None:
                self.book_id = series_match[4]
                self.title = series_match[3]
                self.author = series_match[1]

            self.series = [
                series_match[1]
                for series_match in self._series_link_pattern.finditer(self.content)
            ]

    def render(self) -> None:
        """Render markdown file content."""
        required = ["book_id", "title", "author", "tags", "series", "review", "rating"]
        for attribute in required:
            if getattr(self, attribute) is None:
                raise ValueError(f"To create review file need attribute '{attribute}'")
        assert self.tags is not None  # to please mypy
        assert self.author is not None  # to please mypy
        assert self.series is not None  # to please mypy
        if "#book/book" not in self.tags:
            self.tags.append("#book/book")
        if self.rating is not None and self.rating > 0:
            rating_tag = f"#book/rating{self.rating}"
            if rating_tag not in self.tags:
                self.tags.append(rating_tag)
        book_url = f"https://www.goodreads.com/book/show/{self.book_id}"
        self.content = f"""
[[{clean_file_name(self.author)}]]: [{self.title}]({book_url})
ISBN{self.isbn} (ISBN13{self.isbn13})
{' '.join(['[[' + self.series_full_name(series) + ']]' for series in self.series])}
{self.review}

[Search in Calibre](calibre://search/_?q={urllib.parse.quote(self.title)})

{" ".join(self.tags)}
"""

    @property  # type: ignore
    def file_name(self) -> str:
        """Markdown file name.

        Automatically generate file name from book's fields if not assigned.
        """
        if self._file_name is None:
            assert self.title is not None
            assert self.author is not None
            self._file_name = f"{clean_file_name(self.author)} - {clean_file_name(self.title)}.md"
        return self._file_name

    def series_full_name(self, series: str) -> str:
        """Return file name without extension for series."""
        assert self.author is not None
        return clean_file_name(f"{self.author} - {series} - series")

    def is_series_file_name(self) -> bool:
        """Return True if file name if indicate this is series description file."""
        assert self.file_name is not None
        return re.match(r".* - .* - series\.md$", self.file_name) is not None

    def series_file_name(self, series: str) -> str:
        """Return file name for series."""
        return f"{self.series_full_name(series)}.md"

    @file_name.setter  # type: ignore  # same name as property
    def file_name(self, file_name: str) -> None:
        """Set file_name.

        Set None by default (if not in __init__() params)
        """
        if isinstance(file_name, property):
            self._file_name = None
            return
        self._file_name = file_name

    def write(self) -> None:
        """Write markdown file to path.

        If the file with `author - title.md` already exists, append book ID to the file name.
        Even if this file also exists that does not matter because this it the book
        with the same ID.
        """
        assert self.file_name is not None  # to please mypy
        assert self.content is not None  # to please mypy
        if (
            (self.folder / self.file_name).exists()
            and self.book_id is not None
            and self.book_id not in self.file_name
        ):
            self.file_name = f"{self.file_name} - {self.book_id}.md"
        with (self.folder / self.file_name).open("w", encoding="utf8") as file:
            file.write(self.content)

    def delete_file(self) -> None:
        """Delete the book file."""
        assert (
            self.folder is not None and self.file_name is not None
        ), "Can not delete file without folder"
        if (self.folder / self.file_name).exists():
            os.remove(self.folder / self.file_name)

    def series_list_full_names(self) -> Dict[str, str]:
        """Return dict of series full names."""
        assert self.series is not None  # to please mypy
        return {series: self.series_full_name(series) for series in self.series}

    def delete_series_files(self) -> Dict[str, Path]:
        """Delete series files for review.

        Return deleted series files {series name: series file path}
        """
        deleted_series_files = {}
        assert self.series is not None
        for series, series_full_name in self.series_list_full_names().items():
            series_file_path = self.folder / f"{series_full_name}.md"
            if series_file_path.exists():
                os.remove(series_file_path)
                deleted_series_files[series] = series_file_path
        return deleted_series_files

    def rename_author(self, new_author: str) -> Tuple[Dict[str, Path], Dict[str, Path]]:
        """Rename author in review file.

        In content replace links only, do not re-render content to keep user changes.
        Also recreate series files with new author name in the file names.

        Return (deleted series files, created series files)
        """
        assert self.author is not None
        assert self.content is not None
        old_series_names = self.series_list_full_names()
        deleted_series_files = self.delete_series_files()
        self.delete_file()
        self.content = self.content.replace(f"[[{self.author}]]", f"[[{new_author}]]")
        self.author = new_author
        self._file_name = None  # to recreate from fields
        for series_name, series_full in old_series_names.items():
            self.content = self.content.replace(series_full, self.series_full_name(series_name))
        created_series_files = self.create_series_files()
        self.write()
        return deleted_series_files, created_series_files

    def create_series_files(self) -> Dict[str, Path]:
        """Create series files if they do not exist.

        Do not change already existed files.
        Return created series files {series name: series file path}
        """
        assert self.series is not None
        assert self.author is not None
        created_series_files = {}
        for series in self.series:
            series_file_name = self.series_file_name(series)
            search_params = urllib.parse.urlencode(
                {
                    "utf8": "✓",
                    "q": f"{series}, #",
                    "search_type": "books",
                    "search[field]": "title",
                }
            )
            series_file_path = self.folder / series_file_name
            if not series_file_path.is_file():
                with series_file_path.open("w", encoding="utf8") as md_file:
                    md_file.write(
                        f"""[[{clean_file_name(self.author)}]]
[{series}](https://www.goodreads.com/search?{search_params})"""
                    )
                created_series_files[series] = series_file_path
        return created_series_files



"""Create md-files from https://www.goodreads.com/ CSV export."""
import os
import sys
from pathlib import Path

import click

from goodreads_export.goodreads_book import GoodreadsBooks
from goodreads_export.log import Log
from goodreads_export.markdown_book import BooksFolder
from goodreads_export.version import VERSION

GOODREAD_EXPORT_FILE_NAME = "goodreads_library_export.csv"


@click.command()
@click.argument(
    "csv_file",
    default=GOODREAD_EXPORT_FILE_NAME,
    nargs=1,
)
@click.option(
    "--out",
    "-o",
    "output_folder",
    default=".",
    type=click.Path(exists=True, path_type=Path),
    help="Folder where we put result. By default current folder.",
    nargs=1,
)
@click.option(
    "--version",
    "version",
    is_flag=True,
    default=False,
    help="Show version.",
    nargs=1,
)
@click.option(
    "--merge",
    "-m",
    "merge",
    is_flag=True,
    default=False,
    help="""Merge only. Use it if you need only re-link to primary author names
without importing goodreads file. See https://andgineer.github.io/goodreads-export/ for details.""",
    nargs=1,
)
@click.option(
    "--verbose",
    "-v",
    "verbose",
    is_flag=True,
    default=False,
    help="""Increase verbosity.""",
    nargs=1,
)
def main(csv_file: str, output_folder: Path, version: bool, merge: bool, verbose: bool) -> None:
    """Convert reviews and authors from goodreads export CSV file to markdown files.

    For example you can create nice structure in Obsidian.

    How to create goodreads export see in https://www.goodreads.com/review/import
    In 2022 they declare it to be removed by August,
    but at least at the end of 2022 it still works.

    CSV_FILE: Goodreads export file. By default `goodreads_library_export.csv`.
    if you specify just folder it will look for file with this name in that folder.

    Documentation https://andgineer.github.io/goodreads-export/
    """
    try:
        log = Log(verbose)
        if version:
            print(f"{VERSION}")
            sys.exit(0)

        if merge:  # merge only without import from goodreads file
            books = GoodreadsBooks(None)
        else:
            if os.path.isdir(
                csv_file
            ):  # if folder as csv_file try to find goodreads file in that folder
                csv_file = os.path.join(csv_file, GOODREAD_EXPORT_FILE_NAME)
            if not os.path.isfile(csv_file):
                print(f"Goodreads export file '{csv_file}' not found.")
                sys.exit(1)
            log.start(f"Loading reviews from {csv_file}")
            books = GoodreadsBooks(csv_file)
            print(f" loaded {len(books)} reviews.")

        books_folder = BooksFolder(output_folder, log)
        log.start(f"Reading existing files from {output_folder}")
        print(
            f" loaded {len(books_folder.reviews)} books, {len(books_folder.authors)} authors, "
            f"skipped {books_folder.stat.skipped_unknown_files} unknown files"
            f" and {books_folder.stat.series_added} series files.",
        )
        books_folder.merge_author_names()
        books_folder.dump(books)
        print(
            f"\nAdded {books_folder.stat.reviews_added} review files, "
            f"{books_folder.stat.authors_added} author files.",
            f"Renamed {books_folder.stat.authors_renamed} authors, "
            f"removed duplicate {len(books_folder.stat.author_removed_names)} author names.",
        )

    except Exception as exc:  # pylint: disable=broad-except
        print(f"\n{exc}")
        sys.exit(1)


if __name__ == "__main__":
    main()  # pylint: disable=no-value-for-parameter



"""Create markdown files for book review and author."""
import os
from pathlib import Path
from typing import Dict, Tuple

from goodreads_export.author_file import AuthorFile
from goodreads_export.book_file import BookFile
from goodreads_export.goodreads_book import Book, GoodreadsBooks
from goodreads_export.log import Log
from goodreads_export.stat import Stat

SUBFOLDERS = {
    "toread": "toread",  # for books without review and rating - supposedly this is from to-read
    "reviews": "reviews",  # all other books
    "authors": "authors",  # book authors
}

REVIEWS_SUBFOLDERS = [SUBFOLDERS["reviews"], SUBFOLDERS["toread"]]


class BooksFolder:
    """Create markdown files for book review and author."""

    stat = Stat()

    def __init__(self, folder: Path, log: Log) -> None:
        """Initialize."""
        self.folder = folder
        self.log = log
        self.reviews: Dict[str, BookFile] = {}
        for reviews_subfolder in REVIEWS_SUBFOLDERS:
            self.reviews |= self.load_reviews(folder / reviews_subfolder)
        self.authors, self.primary_authors = self.load_authors(folder / SUBFOLDERS["authors"])

    def merge_author_names(self) -> None:
        """Replace all known versions of author names (translations, misspellings) with one primary name.

        All author name versions should be listed as links in one author file -
        just copy them from other author files to that `primary` author file.
        Reviews will be relinked to this file.
        Author files of this author with one name in them will be deleted.
        Also recreate series files with new author name in the file names.
        """
        for review in self.reviews.values():
            if (
                review.author in self.authors
                and review.author != self.authors[review.author].author
            ):
                self.log.debug(  # should be before rename to log old author name
                    f"Modified review {review.file_name}: renamed author {review.author}"
                    f" to {self.authors[review.author].author}"
                )
                deleted_series_files, created_series_files = review.rename_author(
                    self.authors[review.author].author
                )
                self.log.debug(
                    f"Deleted series files: {[str(path) for path in deleted_series_files.values()]}, "
                    f"created: {[str(path) for path in created_series_files.values()]}."
                )
                self.stat.authors_renamed += 1
        for author in self.primary_authors.values():
            removed_names = author.remove_non_primary_files()
            self.log.debug(f"Removed {author.author}'s duplicate names: {removed_names}")
            self.stat.author_removed_names += removed_names

    def dump(self, books: GoodreadsBooks) -> None:
        """Save books and authors as md-files."""
        for subfolder in SUBFOLDERS.values():
            os.makedirs(self.folder / subfolder, exist_ok=True)

        reviews_bar_title = "Review"
        authors_bar_title = "Author"
        self.log.open_progress(reviews_bar_title, "books", len(books))
        self.log.open_progress(authors_bar_title, "authors", bar_format="{desc}: {n_fmt}")

        for book in books:
            self.log.progress(reviews_bar_title)
            self.log.progress_description(reviews_bar_title, f"{book.title}")
            if self.stat.unique_author(book.author):
                self.log.progress(authors_bar_title)
            if book.author in self.authors and book.author != (
                primary_author := self.authors[book.author].author
            ):
                self.log.progress_description(
                    authors_bar_title, f"Author {book.author} changed to {primary_author}"
                )
                book.author = primary_author  # use the same author name for all synonyms

            if book.book_id not in self.reviews:
                if book.author not in self.authors and self.create_author_md(book):
                    self.stat.authors_added += 1
                    self.log.progress_description(authors_bar_title, f"Added author {book.author}")
                added_file_path = self.create_book_md(book)
                # we know there was no file with this book ID so we added it for sure
                self.stat.reviews_added += 1
                self.log.debug(f"Added review {book.title}, {added_file_path} ")
        self.log.close_progress()

    def create_book_md(self, book: Book) -> str:
        """Create book markdown file.

        Return the filename.
        """
        if book.review == "" and book.rating == 0:
            subfolder = SUBFOLDERS["toread"]
        else:
            subfolder = SUBFOLDERS["reviews"]
        book_markdown = BookFile(
            title=book.title,
            folder=self.folder / subfolder,
            tags=book.tags,
            author=self.authors[book.author].author
            if book.author in self.authors
            else book.author,
            book_id=book.book_id,
            rating=book.rating,
            isbn=book.isbn,
            isbn13=book.isbn13,
            series=book.series,
            review=book.review,
        )
        book_markdown.create_series_files()
        book_markdown.write()
        assert book_markdown.file_name is not None  # to make mypy happy
        return os.path.join(subfolder, book_markdown.file_name)

    def create_author_md(self, book: Book) -> bool:
        """Create author markdown if id does not exist.

        Do not change already existed file.

        Return True if author file was added, False otherwise
        """
        author_markdown = AuthorFile(
            author=book.author,
            folder=self.folder / SUBFOLDERS["authors"],
        )
        assert author_markdown.file_name is not None  # to make mypy happy
        if not (author_markdown.folder / author_markdown.file_name).is_file():
            author_markdown.write()
            return True
        return False

    def load_reviews(self, folder: Path) -> Dict[str, BookFile]:
        """Load existed book reviews.

        Look for goodreads book ID inside files.
        Return {id: BookFile} for files with book ID, ignore other files.
        This way we ignore "- series" files and unknown files.
        """
        reviews: Dict[str, BookFile] = {}
        for file_name in folder.glob("*.md"):
            with file_name.open("r", encoding="utf8") as review_file:
                book = BookFile(
                    title=Path(file_name).stem,
                    folder=folder,
                    file_name=file_name.name,
                    content=review_file.read(),
                )
                if book.book_id is None:
                    if book.is_series_file_name():
                        self.stat.series_added += 1
                    else:
                        self.stat.skipped_unknown_files += 1
                elif book.book_id in reviews:
                    raise ValueError(
                        f"Duplicate book ID {book.book_id} in {file_name} "
                        f"and {reviews[book.book_id].file_name}"
                    )
                else:
                    reviews[book.book_id] = book
        return reviews

    def load_authors(self, folder: Path) -> Tuple[Dict[str, AuthorFile], Dict[str, AuthorFile]]:
        """Load existed authors.

        Look for author synonyms inside files and connect them with `Primary` author file - file
        with multiple author links inside.
        Create dict with {author: AuthorFile}, for each synonym.
        Return (all-authors, primary-authors)
        """
        authors: Dict[str, AuthorFile] = {}
        primary_authors: Dict[str, AuthorFile] = {}
        for file_name in folder.glob("*.md"):
            with file_name.open("r", encoding="utf8") as author_file:
                author = AuthorFile(
                    folder=folder,
                    file_name=file_name.name,
                    author=Path(file_name).stem,
                    content=author_file.read(),
                )
                assert author.names is not None  # to make mypy happy
                if len(author.names):
                    primary_file = (
                        len(author.names) > 1
                    )  # relink to primary file - with author names list
                    if primary_file:
                        primary_authors[author.author] = author
                    for name in author.names + [author.author]:  # + [author.author] to add
                        # `author` created from the file name, it's not necessary in the links inside the file
                        if name in authors and primary_file or name not in authors:
                            assert (
                                name not in authors
                                or author.author == authors[name].author
                                or len(authors[name].names) < 2  # type: ignore
                            ), (
                                f"Multiple author files `{author.author}` and `{authors[name].author}`"
                                " with name versions - should be only one `Primary` file"
                                " with multiple name versions."
                            )
                            authors[name] = author
                else:
                    self.stat.skipped_unknown_files += 1
        return authors, primary_authors



"""Loading statistics."""
from typing import List, Set


class Stat:  # pylint: disable=too-few-public-methods
    """Loading statistics."""

    reviews_added: int = 0
    authors_added: int = 0
    skipped_unknown_files: int = 0
    series_added: int = 0
    authors_renamed: int = 0
    author_removed_names: List[str] = []
    unique_authors: Set[str] = set()

    def unique_author(self, author: str) -> bool:
        """Return True if the author is unique."""
        if author not in self.unique_authors:
            self.unique_authors.add(author)
            return True
        return False



"""Author's file."""  # pylint: disable=duplicate-code
import os
import re
import urllib.parse
from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Optional

from goodreads_export.clean_file_name import clean_file_name


@dataclass
class AuthorFile:
    """Author's file.

    On init extract fields from `content` - override other parameters.
    To re-parse the `content` call `parse()`.
    `render()` generate `content` from fields.
    """

    author: str
    folder: Path
    file_name: Optional[str] = None
    _file_name: Optional[str] = field(repr=False, init=False)
    content: Optional[str] = None

    _name_link_pattern: re.Pattern[str] = field(
        repr=False,
        init=False,
        default=re.compile(
            r"\[([^]]*)\]\(https://www\.goodreads\.com/search\?utf8=%E2%9C%93&q=[^&]*"
            r"&search_type=books&search%5Bfield%5D=author\)"
        ),
    )
    names: Optional[list[str]] = field(init=False, default=None)

    def __post_init__(self) -> None:
        """Extract fields from content."""
        self.parse()  # we do not run parse on content assign during __init__()
        if self.content is None:
            self.render()

    def parse(self) -> None:
        """Parse markdown file content."""
        if self.content is not None:
            self.names = [match[1] for match in self._name_link_pattern.finditer(self.content)]
        elif self.names is None:
            self.names = [self.author]

    def render(self) -> None:
        """Render markdown file content."""
        search_params = urllib.parse.urlencode(
            {
                "utf8": "✓",
                "q": self.author,
                "search_type": "books",
                "search[field]": "author",
            }
        )
        self.content = f"""[{self.author}](https://www.goodreads.com/search?{search_params})

#book/author
"""

    @property  # type: ignore
    def file_name(self) -> str:
        """Markdown file name.

        Automatically generate file name from book's fields if not assigned.
        """
        if self._file_name is None:
            self._file_name = f"{clean_file_name(self.author)}.md"
        return self._file_name

    @file_name.setter
    def file_name(self, file_name: str) -> None:
        """Set file_name.

        Set None by default (if not in __init__() params)
        """
        if isinstance(file_name, property):
            self._file_name = None
            return
        self._file_name = file_name

    def remove_non_primary_files(self) -> List[str]:
        """Remove files with non-primary author names.

        Return list of removed names.
        """
        assert self.names is not None  # to make mypy happy
        removed_names = []
        for name in self.names:
            if (
                name != self.author
                and (author_file_path := self.folder / f"{clean_file_name(name)}.md").exists()
            ):
                os.remove(author_file_path)
                removed_names += [name]
        return removed_names

    def write(self) -> None:
        """Write markdown file to path."""
        assert self.file_name is not None  # to please mypy
        assert self.content is not None  # to please mypy
        with (self.folder / self.file_name).open("w", encoding="utf8") as file:
            file.write(self.content)
